@@ -17,7 +17,10 @@ print "ok 1\n";
 # Insert your test code below (better if it prints "ok 13"
 # (correspondingly "not ok 13") depending on the success of chunk 13
 # of the test code):
-my $openssl = new OpenCA::OpenSSL( SHELL => "/usr/bin/openssl");
+my $openpath=`type -path openssl`;
+chomp $openpath;
+
+my $openssl = new OpenCA::OpenSSL( SHELL => $openpath);
 
 my $k = {
 	CONFIG => "/usr/ssl/openssl.cnf"
